/*
 * restrict_thread_drv driver -
 *
 *  Copyright 2010 Sony Corporation.
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#include <linux/module.h>
#include <linux/kernel.h>

#include <linux/fs.h>


#include <linux/interrupt.h>
#include <asm/bitops.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <asm/semaphore.h>

#include <asm-generic/cputime.h>
#include <linux/kthread.h>

#include "restrict_thread_drv.h"

static int s_rest_thd_debug = 0;

#define REST_THD_DP0(fmt, args ...) ( s_rest_thd_debug >= 0 ? printk("RestThd "fmt, ##args) : 0 )
#define REST_THD_DP1(fmt, args ...) ( s_rest_thd_debug >= 1 ? printk("RestThd "fmt, ##args) : 0 )
#define REST_THD_DP2(fmt, args ...) ( s_rest_thd_debug >= 2 ? printk("RestThd "fmt, ##args) : 0 )
#define REST_THD_DP3(fmt, args ...) ( s_rest_thd_debug >= 3 ? printk("RestThd "fmt, ##args) : 0 )


#define DEF_REST_THD_DEFAULT_ACTIVE_LIMIT_JIF            10
#define DEF_REST_THD_DEFAULT_RESTRICT_RESET_JIF          100
#define DEF_REST_THD_DEFAULT_SEARCH_THREAD_GROUP_CNT     10 // multi RESTRICT_RESET_JIF  

#define DEF_REST_THD_MAX_ACTIVE_CNT 200 //enable to regist 200 thread.

#define DEF_REST_THD_WAKEUP_JIF 1 //wakeup every 1 jiffiles

#define DEF_REST_THD_DEBUG_MEAGURE_CYCLE 2500 //10sec

#define DEF_REST_THD_DEFAULT_SET_PRIORITY 19
#define DEF_REST_THD_DEFAULT_SET_POLICY SCHED_RR

#define DEF_REST_THD_DEFAULT_SCHED_THD_PRIORITY 61

#define DEF_REST_THD_DEFAULT_RESTRICT_THD_PRIORITY   0
#define DEF_REST_THD_DEFAULT_RESTRICT_THD_POLICY     SCHED_NORMAL

#define DEF_REST_THD_DEFAULT_DEBUG_INVALID_PRIORITY    -1

static restrict_thread_info_t g_rest_thd_info;

static void _restrict_thread_timer_fn( unsigned long arg );
static void _debug_trace_cycle_info( restrict_thread_info_t *p_info );
static void _debug_print_active_list( restrict_thread_info_t *p_info );
static void _debug_print_restrict_wait_list( restrict_thread_info_t *p_info );
static void _debug_print_restrict_done_list( restrict_thread_info_t *p_info );
static void _debug_print_all_thread_info( restrict_thread_info_t *p_info );
static void _debug_thread_in_thread_group( restrict_thread_info_t *p_info );

static active_thread_entry_t* _allocate_active_thread_entry( restrict_thread_info_t *p_info )
{
    return kmem_cache_alloc( p_info->active_cache, GFP_ATOMIC );
}

static void _free_active_thread_entry( restrict_thread_info_t *p_info, active_thread_entry_t *p_entry )
{
    kmem_cache_free( p_info->active_cache, p_entry );
}

static pid_entry_t* _allocate_pid_entry( restrict_thread_info_t *p_info )
{
    return kmem_cache_alloc( p_info->pid_entry_cache, GFP_ATOMIC );
}

static void _free_pid_entry( restrict_thread_info_t *p_info, pid_entry_t *p_entry )
{
    kmem_cache_free( p_info->pid_entry_cache, p_entry );
}

static void _restrict_thread_add_timer( restrict_thread_info_t *p_info )
{
    p_info->rest_timer_list.expires = jiffies + DEF_REST_THD_WAKEUP_JIF;
    add_timer( &p_info->rest_timer_list );

}

/* set timer handler*/
static void _start_restrict_thread_timer( restrict_thread_info_t *p_info ) 
                                 
{
    set_bit( DEF_REST_THD_START_TIMER, &p_info->rest_thread_flag );

    /*next restrict jif = now jiffies + restrict_reset_jif*/
    p_info->next_restrict_reset_jif = (long)jiffies + (long)p_info->rest_time.restrict_reset_jif;
    p_info->next_search_thread_group_jif = (long)jiffies + (long)p_info->search_thread_group_jif;

    p_info->debug_total_reset_cnt = 0;
    p_info->debug_total_restrict_cnt = 0;

    p_info->rest_timer_list.data = (unsigned long)p_info;
    p_info->rest_timer_list.function = _restrict_thread_timer_fn;

    _restrict_thread_add_timer( p_info );
}

/* delete timer handler*/
void _stop_restrict_thread_timer( restrict_thread_info_t *p_info )                                  
{
    clear_bit( DEF_REST_THD_START_TIMER, &p_info->rest_thread_flag );

    del_timer_sync( &p_info->rest_timer_list );
}


/* kthread_run handler*/
static void _wakeup_sched_thread( sched_thd_info_t *p_sched_info )
{
    set_bit( DEF_REST_THD_SCHED_THD_WAKEUP, &p_sched_info->flag );
    if( waitqueue_active( &p_sched_info->wait_queue ) ){
        wake_up_interruptible( &p_sched_info->wait_queue );
    }
}

/*Call sched_setscheduler() , If p_task is not NULL.*/
static int _call_sched_setscheduler( int policy,
                                     struct sched_param* p_sched_param,
                                     pid_t pid  )
{
    int ret = 0;
    struct task_struct *p_task;

    read_lock(&tasklist_lock);
    p_task = find_task_by_pid( pid );
    if( p_task ){
        sched_setscheduler(p_task, policy, p_sched_param);
        REST_THD_DP1("<%s>change schedule pid=%d policy=%d prio=%d\n",__func__,
                     pid,policy,(int)p_sched_param->sched_priority);
    } else {
        REST_THD_DP1("<%s>p_task is null \n",__func__);
        ret = -EFAULT;
    }
	read_unlock(&tasklist_lock);

    return ret;
}

static int _set_priority_and_call_setscheduler( restrict_thread_info_t *p_info,
                                       rest_thd_info_t *p_thd_info )
{

    /*set priority from user*/
    p_info->set_sched_param.sched_priority = p_thd_info->priority;
    if( p_thd_info->priority != 0 ){
        p_info->set_policy = SCHED_RR;
    } else {
        p_info->set_policy = SCHED_NORMAL;
    }

    return _call_sched_setscheduler( p_info->set_policy, &p_info->set_sched_param,
                                     p_thd_info->pid );
}

/*Is search thread group over*/
static int _is_search_thread_group( restrict_thread_info_t *p_info )
{
    int ret = 0;    
    if( time_after_eq( jiffies, p_info->next_search_thread_group_jif ) ){
        ret = 1;
    }
    return ret;
}

/* Is reset restrict thread over? */
static int _is_reset_restrict_thread( restrict_thread_info_t *p_info )
{
    int ret = 0;
    if( time_after_eq( jiffies, p_info->next_restrict_reset_jif ) ){
        ret = 1;
    }
    return ret;
}

/*If current is in parent's thread group, return 1*/
static int _is_current_thread_in_group_thread( restrict_thread_info_t *p_info )
{
    return (current->tgid == p_info->parent_pid) ? 1 : 0;
}

static void _init_active_thread_info( active_thread_info_t *p_active_info )
{
    p_active_info->pid = current->pid;
    p_active_info->total_active_time = 1;
    p_active_info->status = 0;
}

static active_thread_entry_t *_allocate_and_init_active_thread_entry( restrict_thread_info_t *p_info )
{
    active_thread_entry_t *p_active_entry = _allocate_active_thread_entry( p_info );
    if( p_active_entry ){
        _init_active_thread_info( &p_active_entry->active_thread_info );
    }
    return p_active_entry;
}

static int _add_active_list( restrict_thread_info_t *p_info )
{
    int ret = 0;
    unsigned long flags;
    active_thread_entry_t *p_active_entry = _allocate_and_init_active_thread_entry( p_info );
    
    if( p_active_entry ){
        /*This list is touched from thread and timer_handler.
         So it is necessary to lock.*/
        spin_lock_irqsave( &p_info->active_list_lock, flags );
        list_add_tail( &p_active_entry->list, &p_info->active_list );
        spin_unlock_irqrestore( &p_info->active_list_lock, flags );

        atomic_inc( &p_info->active_cnt );
    } else {
        ret = -ENOMEM;
    }

    return ret;
}

static void _init_pid_entry( pid_entry_t *p_restrict_entry, pid_t pid )
{
    p_restrict_entry->pid = pid;
}

static pid_entry_t *_allocate_and_init_pid_entry( restrict_thread_info_t *p_info, pid_t pid )
{
    pid_entry_t *p_pid_entry = _allocate_pid_entry( p_info );
    if( p_pid_entry ){
        _init_pid_entry( p_pid_entry, pid );
    }
    return p_pid_entry;
}

/* allocate pid entry and add restrict_wait_list*/
static int _add_restrict_wait_list( restrict_thread_info_t *p_info )
{
    int ret = 0;
    unsigned long flags;
    pid_entry_t *p_pid_entry = _allocate_and_init_pid_entry( p_info, current->pid );

    if( p_pid_entry ){
        spin_lock_irqsave( &p_info->restrict_wait_list_lock, flags );
        list_add_tail( &p_pid_entry->list, &p_info->restrict_wait_list );
        spin_unlock_irqrestore( &p_info->restrict_wait_list_lock, flags );

        atomic_inc( &p_info->restrict_wait_cnt );
    } else {
        ret = -ENOMEM;
    }

    return ret;
}

/*delete all entry in active_list*/
static void _delete_all_entry_in_active_list( restrict_thread_info_t *p_info )
{
    unsigned long flags;
    struct list_head *ptr,*next_ptr;
    active_thread_entry_t *entry = NULL;

    spin_lock_irqsave( &p_info->active_list_lock, flags );

    list_for_each_safe( ptr, next_ptr, &p_info->active_list ){
        entry = list_entry( ptr, active_thread_entry_t, list );

        list_del( &entry->list );
        _free_active_thread_entry( p_info, entry );

    }
    spin_unlock_irqrestore( &p_info->active_list_lock, flags );

    atomic_set( &p_info->active_cnt, 0 );
}

/*delete all entry in restrict_wait_list*/
static void _delete_all_entry_in_restrict_wait_list( restrict_thread_info_t *p_info )
{
    unsigned long flags;
    struct list_head *ptr,*next_ptr;
    pid_entry_t *entry = NULL;

    spin_lock_irqsave( &p_info->restrict_wait_list_lock, flags );

    list_for_each_safe( ptr, next_ptr, &p_info->restrict_wait_list ){
        entry = list_entry( ptr, pid_entry_t, list );
        list_del( &entry->list );
        _free_pid_entry( p_info, entry );
    }
    spin_unlock_irqrestore( &p_info->restrict_wait_list_lock, flags );

    atomic_set( &p_info->restrict_wait_cnt, 0 );
}

/*delete all entry in restrict_done_list*/
static void _delete_all_entry_in_restrict_done_list( restrict_thread_info_t *p_info )
{
    unsigned long flags;
    struct list_head *ptr,*next_ptr;
    pid_entry_t *entry = NULL;

    spin_lock_irqsave( &p_info->restrict_done_list_lock, flags );

    list_for_each_safe( ptr, next_ptr, &p_info->restrict_done_list ){
        entry = list_entry( ptr, pid_entry_t, list );
        list_del( &entry->list );
        _free_pid_entry( p_info, entry );
    }
    spin_unlock_irqrestore( &p_info->restrict_done_list_lock, flags );

    atomic_set( &p_info->restrict_done_cnt, 0 );
}

/*delete entry from active_list, restrict_wait_list, and restrict_done_list*/
static void _delete_entry_from_all_list( restrict_thread_info_t *p_info )
{
    _delete_all_entry_in_active_list( p_info );
    _delete_all_entry_in_restrict_wait_list( p_info );
    _delete_all_entry_in_restrict_done_list( p_info );
}


/*This function is called by timer_handler.
  add own active time and if active time is over limit , 
  set_bit and return 1*/
static int _check_and_update_active_time( restrict_thread_info_t *p_info,
                                          active_thread_info_t *p_active_info)
{
    int need_wakeup = 0;

    /*This thread has already been restricted, do nothing*/
    if( test_bit( DEF_REST_THD_STATUS_ALREADY_RESTRICT, &p_active_info->status ) ){
        return need_wakeup;
    }
    
    p_active_info->total_active_time++;
    REST_THD_DP3(" %s total:%d limit:%d\n",__func__,
                 (int)p_active_info->total_active_time,
                 (int)p_info->rest_time.active_limit_jif );

    if( p_active_info->total_active_time >= p_info->rest_time.active_limit_jif ){
        need_wakeup = 1;        
        set_bit( DEF_REST_THD_STATUS_ALREADY_RESTRICT, &p_active_info->status );
    }
    return need_wakeup;
}

/*add total time. 
  First, If current is in the parent's thread group , add active_time.
  Second If current's active_time is over, add entry in restrict_wait_list
         and wakeup for sched thread.
  Third  If current is not in active list, add it. */  
static int _update_active_and_add_list( restrict_thread_info_t *p_info )
{
    int need_wakeup = 0;
    int is_pid_exist = 0;
    unsigned long flags;
    struct list_head *ptr;
    active_thread_entry_t *entry = NULL;

    /*current is not the same thread group for parent */
    if( !_is_current_thread_in_group_thread( p_info ) ){
        return need_wakeup;
    }

    /*search current from add list*/
    spin_lock_irqsave( &p_info->active_list_lock, flags );

    list_for_each( ptr, &p_info->active_list ){
        entry = list_entry( ptr, active_thread_entry_t, list );
        
        /*current is in active list*/
        if( current->pid == entry->active_thread_info.pid ){

            /* If current is founded in active list*/
            is_pid_exist = 1;

            /* add total time. If active time is over , need_wakeup is 1. */
            need_wakeup = _check_and_update_active_time( p_info, &entry->active_thread_info );

            if( need_wakeup ){
                /*this p_task is used debug only. So do not lock*/
                struct task_struct *p_task = find_task_by_pid( entry->active_thread_info.pid );
                if( p_task ){
                    REST_THD_DP2("update jif[now:0x%08x next_r:0x%08x next_s:0x%08x] pid:%d status:%d total_jif:%d  policy:%d prio:%d wakeup:%d\n",
                                (unsigned int)jiffies,
                                (unsigned int)p_info->next_restrict_reset_jif,
                                (unsigned int)p_info->next_search_thread_group_jif,
                                (int)entry->active_thread_info.pid,
                                (int)entry->active_thread_info.status,
                                (int)entry->active_thread_info.total_active_time,
                                (int)p_task->policy,
                                (int)p_task->rt_priority,
                                need_wakeup );
                } else {
                    REST_THD_DP2("update jif[now:0x%08x next_r:0x%08x next_s:%d] pid:%d status:%d total_jif:%d  wakeup:%d\n",
                                (unsigned int)jiffies,
                                (unsigned int)p_info->next_restrict_reset_jif,
                                (unsigned int)p_info->next_search_thread_group_jif,
                                (int)entry->active_thread_info.pid,
                                (int)entry->active_thread_info.status,
                                (int)entry->active_thread_info.total_active_time,
                                need_wakeup );
                }
            }
            break;
        }
    }

    spin_unlock_irqrestore( &p_info->active_list_lock, flags );

    /*If there is a thread to restrict. */
    if( need_wakeup ){
        /*add restrict wait list*/
        if( _add_restrict_wait_list( p_info ) != 0 ){
            REST_THD_DP0("memory empty allocate pid_entry_t \n");
            need_wakeup = 0;
        }
    }

    /* current is not in the list. So add*/
    if( !is_pid_exist ){
        if( atomic_read( &p_info->active_cnt ) <= DEF_REST_THD_MAX_ACTIVE_CNT ){
            /* add list*/
            _add_active_list( p_info );
        } else {
            printk("too many thread in restrict thread cnt=%d current_pid=%d\n",
                   atomic_read( &p_info->active_cnt ),
                   (int)current->pid );
        }
    }
    
    return need_wakeup;
}


/*timer handler
  1. check search_threa_group. If time is next_search_thread_group_jif, 
     set bit and wakeup sched thread.
  2. check restrict_reset. If time is next_restrict_reset_jif,
     set bit and wakeup sched thread.
  3. update active time. If current is in parent's thread group, add 
     active time. And If active time is over limit , set bit and
     wakeup sched thread. 
*/
static void _restrict_thread_timer_fn( unsigned long arg )
{
    int need_wakeup = 0;
    restrict_thread_info_t *p_info = ( restrict_thread_info_t * )arg;

    /*If user stoped, */
    if( test_bit( DEF_REST_THD_STOP_REST_THREAD, &p_info->rest_thread_flag ) ){
        return;
    }

    _debug_trace_cycle_info( p_info );

    /* First , check search thread group .If elapsed  , search all thread in parent's 
       thread group and change priority setting user.*/
    if( _is_search_thread_group( p_info ) ){

        REST_THD_DP2("Search pid(%d)'s thread \n",p_info->parent_pid );
        set_bit( DEF_REST_THD_STATE_FLAG_SEARCH_THREAD_GROUP, &p_info->sched_thread_info.state_flag);

        /*set next restrict and search time*/
        p_info->next_search_thread_group_jif = jiffies + (long)p_info->search_thread_group_jif;
        p_info->next_restrict_reset_jif = (long)jiffies + (long)p_info->rest_time.restrict_reset_jif;
        need_wakeup = 1;

    } else {
        /*If set no restrict mode ( set prioroty only ) skipp following proceed*/
        if( !test_bit( DEF_REST_THD_NO_RESTRICT_MODE, &p_info->rest_thread_flag ) ){

            /* Second, check restrict reset. If elapsed , all thread in restrict_done_list 
               change priority setting user.*/
            if( _is_reset_restrict_thread( p_info ) ){

                REST_THD_DP2("restrict thread jiffies:0x%08x next_r:0x%08x\n",
                             (unsigned int)jiffies,
                             (unsigned int)p_info->next_restrict_reset_jif );
                set_bit( DEF_REST_THD_STATE_FLAG_RESET_THREAD, &p_info->sched_thread_info.state_flag);

                /*set next restrict time*/
                p_info->next_restrict_reset_jif = (long)jiffies + (long)p_info->rest_time.restrict_reset_jif;
                need_wakeup = 1;

            } else {
                /*Third check current thread.If current thread is in parent's thread group,
                  add total_active time , set restrict_wait_thread If  time is over*/
                need_wakeup |= _update_active_and_add_list( p_info );
                if( need_wakeup ){
                    set_bit( DEF_REST_THD_STATE_FLAG_RESTRICT, &p_info->sched_thread_info.state_flag);
                }
            }
        }
    }

    /*if need, wakeup watch thread */
    if( need_wakeup ){
        _wakeup_sched_thread( &p_info->sched_thread_info );
    }

    _restrict_thread_add_timer( p_info );/*set next timer*/
}

/* If parent and thread belonging in parent's thread group are not same setting priority,
   call sched_setscheduler to set priority*/
static void _search_thread_group_and_set_schedululer( restrict_thread_info_t *p_info )
{
    struct task_struct *p_task,*p_next_task;
    
    read_lock(&tasklist_lock);

    p_task = find_task_by_pid( p_info->parent_pid );
    if( (p_task)&&
        (p_task->rt_priority != p_info->set_sched_param.sched_priority) ){
        /*If priority is not same, set scheduler.*/
        _call_sched_setscheduler( p_info->set_policy, &p_info->set_sched_param, p_info->parent_pid );
    }

    p_next_task = p_task;
    if( p_task ){
        /*search parent's thread group*/
        while_each_thread( p_task, p_next_task ){
            if( (p_next_task) &&
                (p_next_task->rt_priority != p_info->set_sched_param.sched_priority) ){
                /*If priority is not same add list*/                
                _call_sched_setscheduler( p_info->set_policy, &p_info->set_sched_param, p_next_task->pid );
            }
        }
    }

	read_unlock(&tasklist_lock);
}

/*call sched_setscheduler to restrict thread.*/
static void _call_sched_setscheduler_thread_restrict( pid_t pid )
{
    struct sched_param _sched_param;
    int policy = DEF_REST_THD_DEFAULT_RESTRICT_THD_POLICY;

    _sched_param.sched_priority = DEF_REST_THD_DEFAULT_RESTRICT_THD_PRIORITY;

    _call_sched_setscheduler( policy, &_sched_param, pid );
}

/*restrict thread in restrict wait list.
 First get thread entry from restrict_wait_thread and restrict thread.
 Second delete entry from restrict_wait_thread and add entry to restrict_done_entry.*/
static void _restrict_thread_in_restrict_wait_list( restrict_thread_info_t *p_info )
{
    unsigned long flags;
    struct list_head *ptr, *next_ptr;
    pid_entry_t *entry = NULL;

    spin_lock_irqsave( &p_info->restrict_wait_list_lock, flags );

    list_for_each_safe( ptr, next_ptr, &p_info->restrict_wait_list ){
        entry = list_entry( ptr, pid_entry_t, list );

        /*restrict thread*/
        _call_sched_setscheduler_thread_restrict( entry->pid );
        p_info->debug_total_restrict_cnt++; //debug

        /*delete entry from restrict_wait_list*/
        list_del_init( &entry->list );

        /*add entry to restrict_done_list.*/
        list_add_tail( &entry->list, &p_info->restrict_done_list );
        atomic_inc( &p_info->restrict_done_cnt );
        
    }
    spin_unlock_irqrestore( &p_info->restrict_wait_list_lock, flags );

    atomic_set( &p_info->restrict_wait_cnt, 0 );
}

/* set priority to the entry restricted.
   entry in restrict_done_list is restricted.This function is 
   set back the priority*/
static void _reset_thread_in_restrict_done_list( restrict_thread_info_t *p_info )
{
    unsigned long flags;
    struct list_head *ptr, *next_ptr;
    pid_entry_t *entry = NULL;

    spin_lock_irqsave( &p_info->restrict_done_list_lock, flags );

    /* set back priority to all the entry in restrict_done_list.
       entrys in restrict_done_list are being restricted.*/

    list_for_each_safe( ptr, next_ptr, &p_info->restrict_done_list ){
        entry = list_entry( ptr, pid_entry_t, list );

        /*set back setting priority. */
        _call_sched_setscheduler( p_info->set_policy, &p_info->set_sched_param, entry->pid );
        
        /*delete entry from restrict_done_list and free entry*/
        list_del( &entry->list );
        _free_pid_entry( p_info, entry );
        
    }
    spin_unlock_irqrestore( &p_info->restrict_done_list_lock, flags );

    atomic_set( &p_info->restrict_done_cnt, 0 );
}

/*kthread's main function*/
static int _sched_thread_start_main( void *arg )
{
    int ret;
    restrict_thread_info_t *p_info = (restrict_thread_info_t *)arg;
    sched_thd_info_t *p_sched_info = &p_info->sched_thread_info;

    set_thread_priority(SCHED_RR, DEF_REST_THD_DEFAULT_SCHED_THD_PRIORITY);

    /*keep itself pid*/
    p_sched_info->pid = current->pid;
    REST_THD_DP2("%s watch_thd_pid=%d\n",__func__,(int)p_sched_info->pid);

    while(1){
        /*If need to change priority. be wakeuped */
        ret = wait_event_interruptible( p_sched_info->wait_queue, test_and_clear_bit( DEF_REST_THD_SCHED_THD_WAKEUP, &p_sched_info->flag ) );
        if( ret < 0 ){
            continue;/*If get suspend*/
        }

        if( test_and_clear_bit(DEF_REST_THD_STATE_FLAG_SEARCH_THREAD_GROUP, &p_info->sched_thread_info.state_flag ) ){
            /*set setting priority in parent's thread group*/

            REST_THD_DP1("_search_thread_group jif=0x%08x\n",(unsigned int)jiffies);
            _search_thread_group_and_set_schedululer( p_info );

            _delete_entry_from_all_list( p_info );

        } else if( test_and_clear_bit(DEF_REST_THD_STATE_FLAG_RESET_THREAD, &p_info->sched_thread_info.state_flag ) ){
            /*set setting priority  in restrict_done_list.*/

            REST_THD_DP2("_reset_thread_in_restrict_done_list jif=0x%08x\n",
                         (unsigned int)jiffies);
            _reset_thread_in_restrict_done_list( p_info );

            _delete_entry_from_all_list( p_info );

        } else if( test_and_clear_bit(DEF_REST_THD_STATE_FLAG_RESTRICT, &p_info->sched_thread_info.state_flag ) ){
            /*restrict thread in restrict_wait_thread.*/

            REST_THD_DP3("_restrict_thread_in_restrict_wait_list jif=0x%08x\n",
                         (unsigned int)jiffies);
            _restrict_thread_in_restrict_wait_list( p_info );

        }
    }
}

/*Set parameter from user, and start timer */
int _set_process_info( restrict_thread_info_t *p_info,
                        rest_thd_info_t *p_thd_info )
{
    int ret;

    /*If setting force stop , do nothing.*/
    if( test_bit( DEF_REST_THD_FORCE_STOP_MODE, &p_info->rest_thread_flag ) ){
        REST_THD_DP0(" set FORCE_STOP_MODE \n");
        return -EFAULT;
    }

    /*If now state is stop, clear bit and do follow sequence*/
    if( !test_and_clear_bit( DEF_REST_THD_STOP_REST_THREAD, &p_info->rest_thread_flag ) ){
        REST_THD_DP0("not stop \n");
        return -EFAULT;
    }

    /*set info from user*/
    p_info->parent_pid = p_thd_info->pid;

    if( test_bit(DEF_REST_THD_SET_DEBUG_PARAM, &p_info->rest_thread_flag ) ){
        //debug set
        REST_THD_DP0("debug on Set debug param prio:%d <time>[limit:%d reset:%d search:%d]\n",
                         p_info->debug_set_priority,
                         (int)p_info->debug_rest_time.active_limit_jif,
                         (int)p_info->debug_rest_time.restrict_reset_jif,
                         (int)p_info->debug_search_jif );
                     
        p_info->rest_time = p_info->debug_rest_time;
        p_info->search_thread_group_jif = p_info->debug_search_jif;
        p_thd_info->priority = p_info->debug_set_priority;
    } else {
        p_info->rest_time = p_thd_info->rest_time;
        p_info->search_thread_group_jif = p_info->rest_time.restrict_reset_jif * DEF_REST_THD_DEFAULT_SEARCH_THREAD_GROUP_CNT;
    }

    /* set priority and policy. And sched_setscheduler*/
    ret = _set_priority_and_call_setscheduler( p_info, p_thd_info );

    REST_THD_DP0("SetParameter pid=%d prio:%d <time>[limit:%d reset:%d search:%d]\n",
                 (int)p_info->parent_pid,
                 p_info->set_sched_param.sched_priority,
                 (int)p_info->rest_time.active_limit_jif,
                 (int)p_info->rest_time.restrict_reset_jif,
                 (int)p_info->search_thread_group_jif );

    if( ret == 0 &&
        !test_bit( DEF_REST_THD_START_TIMER, &p_info->rest_thread_flag ) ){
        /*start timer event*/
        _start_restrict_thread_timer( p_info );
    } else {
        /* If cause some error, set state STOP */
        REST_THD_DP0(" _set_priority_and_call_setscheduler() \n");
        set_bit( DEF_REST_THD_STOP_REST_THREAD, &p_info->rest_thread_flag );
    }

    return ret;
}

/*stop timer */
void _clear_process_info( restrict_thread_info_t *p_info )
{
    set_bit( DEF_REST_THD_STOP_REST_THREAD, &p_info->rest_thread_flag );

    _delete_entry_from_all_list( p_info );
    
    /*stop timer hander if proceed.*/
    if( test_bit( DEF_REST_THD_START_TIMER, &p_info->rest_thread_flag ) ){
        _stop_restrict_thread_timer( p_info );
    }
}

/*debug*/
static void _debug_print_active_list( restrict_thread_info_t *p_info )
{
    unsigned long flags;
    struct list_head *ptr;
    active_thread_entry_t *entry = NULL;

    REST_THD_DP2("\nactive_list(%d)\n",atomic_read( &p_info->active_cnt ) );

    spin_lock_irqsave( &p_info->active_list_lock, flags );

    list_for_each( ptr, &p_info->active_list ){
        entry = list_entry( ptr, active_thread_entry_t, list );
        REST_THD_DP2("pid=%d total=%d status=%x \n",
                     (int)entry->active_thread_info.pid, 
                     (int)entry->active_thread_info.total_active_time,
                     (unsigned int)entry->active_thread_info.status);
        
    }
    spin_unlock_irqrestore( &p_info->active_list_lock, flags );

}

static void _debug_print_restrict_wait_list( restrict_thread_info_t *p_info )
{
    unsigned long flags;
    struct list_head *ptr;
    pid_entry_t *entry = NULL;

    REST_THD_DP2("\nrestrict_wait_list(%d)\n",atomic_read( &p_info->restrict_wait_cnt ) );

    spin_lock_irqsave( &p_info->restrict_wait_list_lock, flags );

    list_for_each( ptr, &p_info->restrict_wait_list ){
        entry = list_entry( ptr, pid_entry_t, list );
        REST_THD_DP2("pid=%d\n",entry->pid );
        
    }
    spin_unlock_irqrestore( &p_info->restrict_wait_list_lock, flags );

}

static void _debug_print_restrict_done_list( restrict_thread_info_t *p_info )
{
    unsigned long flags;
    struct list_head *ptr;
    pid_entry_t *entry = NULL;

    REST_THD_DP2("\nrestrict_done_list(%d)\n",atomic_read( &p_info->restrict_done_cnt ) );

    spin_lock_irqsave( &p_info->restrict_done_list_lock, flags );

    list_for_each( ptr, &p_info->restrict_done_list ){
        entry = list_entry( ptr, pid_entry_t, list );
        REST_THD_DP2("pid=%d\n",entry->pid );
        
    }
    spin_unlock_irqrestore( &p_info->restrict_done_list_lock, flags );

}

static void _debug_print_all_thread_info( restrict_thread_info_t *p_info )
{
    _debug_print_active_list( p_info );
    _debug_print_restrict_wait_list( p_info );
    _debug_print_restrict_done_list( p_info );
}

static void _debug_thread_in_thread_group( restrict_thread_info_t *p_info )
{
    struct task_struct *p_task;
    struct task_struct *p_next_task = p_task;

    read_lock(&tasklist_lock);

    p_task = find_task_by_pid( p_info->parent_pid );
    p_next_task = p_task;

    REST_THD_DP0("Search pid(%d)'s thread \n",p_info->parent_pid );

    if( p_task ){
        REST_THD_DP2("pid=%d pol=%d prio=%d\n",
                     p_task->pid, p_task->policy, p_task->rt_priority);
        while_each_thread( p_task, p_next_task ){
            if( p_next_task ){
                REST_THD_DP2("pid=%d pol=%d prio=%d\n",
                             p_next_task->pid, p_next_task->policy, p_next_task->rt_priority);
            }
            
        }
    }
	read_unlock(&tasklist_lock);
}

static void _debug_trace_cycle_info( restrict_thread_info_t *p_info )
{
    if( p_info->debug_total_reset_cnt >= DEF_REST_THD_DEBUG_MEAGURE_CYCLE ){
        REST_THD_DP1("<%08x> rest:%d restrict=%d <list>active:%d wait:%d done:%d\n",
                     (unsigned int)jiffies,
                     (int)p_info->debug_total_reset_cnt,
                     (int)p_info->debug_total_restrict_cnt,
                     atomic_read(&p_info->active_cnt),
                     atomic_read(&p_info->restrict_wait_cnt),
                     atomic_read(&p_info->restrict_done_cnt) );
        p_info->debug_total_reset_cnt = 0;
        p_info->debug_total_restrict_cnt = 0;

    } else {
        p_info->debug_total_reset_cnt++;
    }
}

int restrict_thread_open( struct inode *inode, struct file *filp )
{
    restrict_thread_info_t *p_info;
    p_info = container_of( inode->i_cdev, restrict_thread_info_t, cdev );

    filp->private_data = p_info;
    return 0;
}

int restrict_thread_release( struct inode *inode, struct file *filp )
{
    return 0;
}

int restrict_thread_ioctl( struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg )
{
    int retval = 0;
    switch( cmd ){
    case REST_THD_SET_INFO:
    {
        /*set parameter from user and start timer*/
        restrict_thread_info_t *p_info = filp->private_data;
        rest_thd_info_t _info;

        if( copy_from_user( &_info, (rest_thd_info_t*)arg, sizeof( rest_thd_info_t ) ) ){
            return -ENOTTY;
        }
        REST_THD_DP0("REST_THD_SET_INFO pid=%d boost=%d range=%d prio=%d\n",
                   _info.pid, (int)_info.rest_time.active_limit_jif,
                   (int)_info.rest_time.restrict_reset_jif, _info.priority);
        retval = _set_process_info( p_info, &_info );
    }
    break;
    case REST_THD_CLEAR_INFO:
    {
        /*clear parameter from user and stop timer*/
        restrict_thread_info_t *p_info = filp->private_data;

        REST_THD_DP0("%s REST_THD_CLEAR_PROCESS_PID 0x%08x\n",
                   __func__,(unsigned int)jiffies);
        _clear_process_info( p_info );
    }
    break;
    case REST_THD_DEBUG_PRINT_LIST:
    {
        /*show thread info in parent's thread group.
          show entry in three list. */
        restrict_thread_info_t *p_info = filp->private_data;
        int t_rest_thd_debug = s_rest_thd_debug;
        s_rest_thd_debug = 3;

        _debug_print_all_thread_info( p_info );
        _debug_thread_in_thread_group( p_info );

        s_rest_thd_debug = t_rest_thd_debug;
    }
    break;
    case REST_THD_SET_LOG_LEVEL:
    {
        if( copy_from_user( &s_rest_thd_debug, (int*)arg, sizeof( int ) ) ){
            return -ENOTTY;
        }
        REST_THD_DP0("Loglevel %d\n",s_rest_thd_debug);
        
    }
    break;
    case REST_THD_SET_MODE:
    {
        restrict_thread_info_t *p_info = filp->private_data;
        int _set_mode;
        if( copy_from_user( &_set_mode, (int*)arg, sizeof( int ) ) ){
            return -ENOTTY;
        }

        REST_THD_DP0("REST_THD_SET_MODE %d\n",_set_mode );

        switch( _set_mode ){
        case DEF_REST_THD_NO_RESTRICT:
            REST_THD_DP0("REST_THD_SET_MODE set=%d\n",_set_mode );
            set_bit( DEF_REST_THD_NO_RESTRICT_MODE, &p_info->rest_thread_flag );
            break;
        case DEF_REST_THD_SET_FORCE_STOP:
            REST_THD_DP0("REST_THD_SET_MODE stop=%d\n",_set_mode );
            set_bit( DEF_REST_THD_FORCE_STOP_MODE, &p_info->rest_thread_flag );
            break;
        case DEF_REST_THD_CLEAR_ALL:
            clear_bit( DEF_REST_THD_NO_RESTRICT, &p_info->rest_thread_flag );
            clear_bit( DEF_REST_THD_FORCE_STOP_MODE, &p_info->rest_thread_flag );
            break;
        default:
            break;
        }
        
    }
    break;
    case REST_THD_SET_DEBUG_PARAM:
    {

        restrict_thread_info_t *p_info = filp->private_data;
        rest_thd_debug_info_t _debug_info;
        if( copy_from_user( &_debug_info, (int*)arg, sizeof( rest_thd_debug_info_t ) ) ){
            return -ENOTTY;
        }
        
        if( _debug_info.type == DEF_REST_THD_DEBUG_TYPE_SET ){
            set_bit( DEF_REST_THD_SET_DEBUG_PARAM, &p_info->rest_thread_flag );
            p_info->debug_set_priority = _debug_info.priority; 
            p_info->debug_rest_time = _debug_info.rest_time; 
            p_info->debug_search_jif = _debug_info.search_thread_group_jif;

            REST_THD_DP0("set debug parameter prio:%d <time>[limit:%d reset:%d search:%d]\n",
                         p_info->debug_set_priority,
                         (int)p_info->debug_rest_time.active_limit_jif,
                         (int)p_info->debug_rest_time.restrict_reset_jif,
                         (int)p_info->debug_search_jif );
        } else {
            clear_bit( DEF_REST_THD_SET_DEBUG_PARAM, &p_info->rest_thread_flag );
            REST_THD_DP0("Clear debug parameter\n");
        }

    }
    break;
    default:
        retval = -ENOTTY;
        }
    return retval;
}


struct file_operations restrict_thread_fops = {
    .owner = THIS_MODULE,
    .open = restrict_thread_open,
    .release = restrict_thread_release,
    .ioctl = restrict_thread_ioctl,
};


static int __init restrict_thread_init( void )
{
    int devno, ret;
    restrict_thread_info_t *p_info;
    p_info = &g_rest_thd_info;

    ret = register_chrdev_region( MKDEV( MAJOR_REST_THD, MINOR_REST_THD ),
                                  1 , DEF_REST_THD_DEV_PATH );

    devno = MKDEV( MAJOR_REST_THD, MINOR_REST_THD );
    cdev_init( &p_info->cdev, &restrict_thread_fops );
    p_info->cdev.owner = THIS_MODULE;
    p_info->cdev.ops = &restrict_thread_fops;
	
    ret = cdev_add( &p_info->cdev, devno, 1 );

    if( ret ){
        printk("test_dev is not ..\n");
        return -1;
    }

    p_info->rest_time.active_limit_jif   = DEF_REST_THD_DEFAULT_ACTIVE_LIMIT_JIF; 
    p_info->rest_time.restrict_reset_jif = DEF_REST_THD_DEFAULT_RESTRICT_RESET_JIF; 
    p_info->search_thread_group_jif = p_info->rest_time.restrict_reset_jif * DEF_REST_THD_DEFAULT_SEARCH_THREAD_GROUP_CNT; 

    INIT_LIST_HEAD( &p_info->active_list );
    spin_lock_init( &p_info->active_list_lock );
    atomic_set( &p_info->active_cnt, 0 );

    INIT_LIST_HEAD( &p_info->restrict_wait_list );
    spin_lock_init( &p_info->restrict_wait_list_lock );
    atomic_set( &p_info->restrict_wait_cnt, 0 );

    INIT_LIST_HEAD( &p_info->restrict_done_list );
    spin_lock_init( &p_info->restrict_done_list_lock );
    atomic_set( &p_info->restrict_wait_cnt, 0 );

    p_info->active_cache = kmem_cache_create( "restrict_tread_active", 
                                              sizeof( active_thread_entry_t ),
                                              0, SLAB_HWCACHE_ALIGN, NULL );
    if( !p_info->active_cache ){
        printk("kmem cache create error\n");
        return -1;
    }

    p_info->pid_entry_cache = kmem_cache_create( "restrict_tread_pid", 
                                                 sizeof( pid_entry_t ),
                                                 0, SLAB_HWCACHE_ALIGN, NULL );
    if( !p_info->pid_entry_cache ){
        printk("kmem cache create error\n");
        return -1;
    }

    //first THD is STOP.
    p_info->rest_thread_flag = 0;
    set_bit( DEF_REST_THD_STOP_REST_THREAD, &p_info->rest_thread_flag );

    p_info->next_restrict_reset_jif = 0;
    p_info->next_search_thread_group_jif = 0;

    p_info->sched_thread_info.state_flag = 0;
    p_info->sched_thread_info.flag = 0;
    init_waitqueue_head(&p_info->sched_thread_info.wait_queue);

    p_info->parent_pid = 0;
    p_info->set_sched_param.sched_priority = DEF_REST_THD_DEFAULT_SET_PRIORITY;
    p_info->set_policy = DEF_REST_THD_DEFAULT_SET_POLICY;

    init_timer( &p_info->rest_timer_list );

    kthread_run( _sched_thread_start_main, p_info, "restrict_thread" );

    //debug
    p_info->debug_total_reset_cnt = 0;
    p_info->debug_total_restrict_cnt = 0;
    p_info->debug_set_priority = DEF_REST_THD_DEFAULT_DEBUG_INVALID_PRIORITY;
    p_info->rest_time.active_limit_jif   = DEF_REST_THD_DEFAULT_ACTIVE_LIMIT_JIF; 
    p_info->rest_time.restrict_reset_jif = DEF_REST_THD_DEFAULT_RESTRICT_RESET_JIF; 
    
    return 0;
}

void __exit restrict_thread_exit(void)
{    
    return;
}

MODULE_LICENSE("GPL");
module_init(restrict_thread_init);
module_exit(restrict_thread_exit);




